"""Semantix is a Python library that give superpowers to your code."""

from semantix.decorators import enhance, tool
from semantix.types.semantic import Semantic, SemanticClass

__all__ = ["Semantic", "enhance", "tool", "SemanticClass"]
